#-----------------------------------------------------------------------------
# Bubbles v1
# 3/28/04
# Copyright (c) 2004 David G. Drumright (ockham)
#-----------------------------------------------------------------------------
# The initially selected object is the 'wand'.  In the script, you pick the
# bubble and set the random and movement parms.  After activation, the
# wand will leave copies of the bubble behind.  The copies will move according
# to the random choices.
#-----------------------------------------------------------------------------

from Tkinter import *
import tkFileDialog
import poser
import random
import os
random.seed()
random.seed()
scene = poser.Scene()
BasePathUncut=poser.AppLocation()
BasePath=BasePathUncut[:(-len('poser.exe'))]
BubblePath=os.path.normpath(BasePath+'/Runtime/Libraries/props/*.PP2')
FPS=float(scene.FramesPerSecond())
SPF=1.0/FPS
TotalSecs=float(scene.NumFrames())/FPS

#-----------------------------------------------------------------------------
# Next section is TK
#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        self.master = master
        master.title("Bubbles")
        self.FileName=0

        #------

        self.StatusEntry = Entry(self.master,width=40)
        self.StatusEntry.grid(row=0,column=0,columnspan=2)

        self.SelectPartButton = Button(self.master, text="Select bubble", command=self.HandleSelect)
        self.SelectPartButton.grid(row=1, column=0)

        self.StartLabel=Label(self.master, text="Start frame:")
        self.StartLabel.grid( row=2,column=0)
        self.StartEntry=Entry(self.master, width=6)
        self.StartEntry.grid(row=2,column=1)

        self.EndLabel=Label(self.master, text="End frame:")
        self.EndLabel.grid( row=3,column=0)
        self.EndEntry=Entry(self.master, width=6)
        self.EndEntry.grid(row=3,column=1)

        #------
        self.SliderFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.SliderFrame.grid(row=4,column=0,sticky=W+E,columnspan=2)
        self.ActionFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ActionFrame.grid(row=5,column=0,sticky=W+E,columnspan=2)
        # Go button.
        self.GoButton = Button(self.master, text="Go", command=self.HandleGo)
        self.GoButton.grid(row=6, column=0)

        #------

        self.EmitSpeedLabel=Label(self.SliderFrame, text="Seconds between emissions:")
        self.EmitSpeedLabel.grid(row=0,column=0,sticky=E)
        self.EmitSpeedSlider=Scale(self.SliderFrame, from_ = SPF, to = TotalSecs,orient=HORIZONTAL,resolution=0.1)
        # Note that this starts and defaults at max possible, namely one per frame.
        self.EmitSpeedSlider.grid(row=0, column=1,columnspan=2)
        self.EmitSpeedSlider.set(SPF)

        self.EmitRandLabel=Label(self.SliderFrame, text="Randomness of emission:")
        self.EmitRandLabel.grid(row=1,column=0,sticky=E)
        self.EmitRandSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        # Note that this starts and defaults at 1 per second.
        self.EmitRandSlider.grid(row=1, column=1,columnspan=2)

        self.RandSizeLabel=Label(self.SliderFrame, text="Randomness of size:")
        self.RandSizeLabel.grid( row=2,column=0,sticky=E)
        self.RandSizeSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandSizeSlider.grid(row=2,column=1,columnspan=2)

        self.RandAngleLabel=Label(self.SliderFrame, text="Randomness of angle:")
        self.RandAngleLabel.grid( row=3,column=0,sticky=E)
        self.RandAngleSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandAngleSlider.grid(row=3,column=1,columnspan=2)


        #------


        self.FallLabel=Label(self.ActionFrame, text="Fall speed:")
        self.FallLabel.grid( row=0,column=0,sticky=E)
        self.FallSlider=Scale(self.ActionFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.FallSlider.grid(row=0,column=1)

        self.DestLabel=Label(self.ActionFrame, text="Dest height:")
        self.DestLabel.grid( row=1,column=0,sticky=E)
        self.DestSlider=Scale(self.ActionFrame, from_ = -2.0, to = 3.0,orient=HORIZONTAL,resolution=.01)
        self.DestSlider.grid(row=1,column=1)

        self.DriftLabel=Label(self.ActionFrame, text="Amount of drift:")
        self.DriftLabel.grid( row=2,column=0,sticky=E)
        self.DriftSlider=Scale(self.ActionFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.DriftSlider.grid(row=2,column=1)

        #------
        # Set the default values for frame entries
        self.StartEntry.insert(0,str(1))
        self.EndEntry.insert(0,str(scene.NumFrames()))

        # Set up the Wand first.
        # The Wand must be the selected item.
        self.WandItem = scene.CurrentActor()

        # If it's part of a figure, get the figure as well.
        self.WandFig = self.WandItem.ItsFigure()
        if (not self.WandItem.IsBodyPart()) and (not self.WandItem.IsProp()):
            self.ShowStatus("Wand prop not a good type.  Cancel and restart.")
        else:
            self.ShowStatus("First select the bubble prop.")

        # Set up close action
        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleSelect(self): # Get the filename for the bubble.
        # Pop up file box to get a Part.
        fn=tkFileDialog.askopenfilename(initialfile=BubblePath, filetypes=[("Props", "PP2")] )
        if not fn:
            self.ShowStatus("No file selected. Try again.")
            return
        self.FileName=os.path.normpath(fn)
        self.ShowStatus("Now set value sliders, then hit Go.")

    # - - - - - - - - - - - - - - - - - -

    def HandleGo(self):
        self.SetupAll()
        self.Fall()
        self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def SetupAll(self):
        if self.FileName==0:
            self.ShowStatus("You need to select a bubble prop first.")
        self.ShowStatus("Working....")
        # Get the sliders we need here.
        EmitSpeed = self.EmitSpeedSlider.get()
        if EmitSpeed < 0.0001: return
        EmitRand= self.EmitRandSlider.get()

        # Figure the interval from the speed.  Speed is seconds per bubble;
        # we need frames per bubble.   (F/S) * (S/B) = F/B.
        EmitInterval = int(float(scene.FramesPerSecond()) * EmitSpeed)
        if EmitInterval < 1: EmitInterval = 1

        RandSize  = float(self.RandSizeSlider.get()) / 100.0
        # For the angle, max move should be 180 degrees.
        RandAngle = float(self.RandAngleSlider.get()) * 180.0 / 100.0

        s = self.StartEntry.get()
        if s:
            StartFrame = int(s)
        s = self.EndEntry.get()
        if s:
            EndFrame = int(s)

        # First loop.
        # Go through the frames, watching the Endpoint of the 'wand'.
        # At intervals given by the EmitSpeed, produce a copy of the 'bubble',
        # at this location, and hold the frame so we can use the second
        # loop to make it appear (scale=1) at the appropriate frame.

        self.BubbleList = {} # Dict: key is name, value is frame.

        # Note that Start and End frame affect the range of creation as well as the
        # range of fall.

        # Always create one copy at StartFrame. After that, we may create several
        # more depending on how the Interval fits.
        F = StartFrame
        while F < EndFrame:  # Loop this way so we can randomize the interval more easily

            scene.SetFrame(F-1)

            scene.DrawAll()

            # Treat the Origin of the WandItem, converted to world coords,
            # as the point of emission.  This lets you pick the exact point
            # conveniently by moving the Origin.
            (x,y,z) = self.WandItem.Origin()
            ( (a,b,c,d) , (e,f,g,h), (i,j,k,l), (m,n,p,q) ) = self.WandItem.WorldMatrix()
            w=1.0
            x1 = x*a + y*e + z*i + w*m
            y1 = x*b + y*f + z*j + w*n
            z1 = x*c + y*g + z*k + w*p
            w1 = x*d + y*h + z*l + w*q
            EmitX = x1/w1
            EmitY = y1/w1
            EmitZ = z1/w1

            # Go back to 0 and make a copy of the Bubble.
            # Reminder: We do a frame-switch here
            # because the dot needs to be at this location from the start, even though
            # we are acting at frame F.
            scene.SetFrame(0)
            scene.LoadLibraryProp(self.FileName)
            # The new prop will be currentactor at this moment.
            NewBubble = scene.CurrentActor()

            # Control its scale randomly. Reminder: the original will be the same
            # every time, but this is easier than trying to pick up a "sample"
            # before the rest.

            YrotParm=NewBubble.ParameterByCode(poser.kParmCodeYROT)
            XrotParm=NewBubble.ParameterByCode(poser.kParmCodeXROT)
            XtranParm=NewBubble.ParameterByCode(poser.kParmCodeXTRAN)
            YtranParm=NewBubble.ParameterByCode(poser.kParmCodeYTRAN)
            ZtranParm=NewBubble.ParameterByCode(poser.kParmCodeZTRAN)

            Yrot=YrotParm.Value() + (random.random()*2.0-1.0)*RandAngle
            Xrot=XrotParm.Value() + (random.random()*2.0-1.0)*RandAngle

            XtranParm.SetValue(EmitX)
            YtranParm.SetValue(EmitY)
            ZtranParm.SetValue(EmitZ)
            YrotParm.SetValue(Yrot)
            XrotParm.SetValue(Xrot)

            # Record the name and the frame when we made it
            self.BubbleList[NewBubble.Name()]=F
            Interval = EmitInterval + (random.random()*2.0-1.0)*EmitRand
            if Interval<1:Interval=1
            F = F + Interval # Reminder: this is a While; will quit when F > numframes.
            # end While

        scene.DrawAll()

        # Second loop: go through the created Bubbles. For each,
        # scale to zero from start, and jump to full scale from the frame where created.
        self.ShowStatus("Yes, I'm still working... Just wait till I close.")
        for NewBubbleName in self.BubbleList.keys():
            ChangeFrame = self.BubbleList[NewBubbleName] # Get the point where it needs to jump (appear)
            NewBubble = scene.Actor(NewBubbleName)
            XscaleParm=NewBubble.ParameterByCode(poser.kParmCodeXSCALE)
            YscaleParm=NewBubble.ParameterByCode(poser.kParmCodeYSCALE)
            ZscaleParm=NewBubble.ParameterByCode(poser.kParmCodeZSCALE)
            AscaleParm=NewBubble.ParameterByCode(poser.kParmCodeASCALE)
            OrigAscale=AscaleParm.Value()
            # Up to ChangeFrame, set scale 0.
            for F in range(1,ChangeFrame): # From 0, not from StartFrame.
                scene.SetFrame(F-1)
                XscaleParm.SetValue(XscaleParm.Value() + (random.random()*2.0-1.0) * RandSize)
                YscaleParm.SetValue(YscaleParm.Value() + (random.random()*2.0-1.0) * RandSize)
                ZscaleParm.SetValue(ZscaleParm.Value() + (random.random()*2.0-1.0) * RandSize)
                AscaleParm.SetValue(0.0)
            # From ChangeFrame to end, set main scale to orig.
            for F in range(ChangeFrame,EndFrame):
                scene.SetFrame(F-1)
                AscaleParm.SetValue(OrigAscale)

        scene.DrawAll()

    # - - - - - - - - - - - - - - - - - -

    def Fall(self): # Acts only if speed has been set
        Speed = float(self.FallSlider.get()) / 1000.0
        if Speed < 0.0001: return  # Zero speed, skip the fall.

        DestY = float(self.DestSlider.get())
        # This may be above or below the original position.
        # If above, we increment Y upward and stop when Y >= DestY.
        # If below, we increment Y dowward and stop when Y <= DestY.
        # To avoid slippery comparisons, we force Y = DestY at the stop point.


        # Fetch drift percent:
        Drift = float(self.DriftSlider.get()) / 500.0
        # No guarding needed here; if 0, it will just naturally
        # leave no random movement.


        StartFrame=1
        EndFrame=scene.NumFrames()
        s = self.StartEntry.get()
        if s:
            StartFrame = int(s)
        s = self.EndEntry.get()
        if s:
            EndFrame = int(s)
        if (EndFrame - StartFrame) < 2: return


        # From startframe to endframe, move each particle down by Increment, and
        # sideways by Drift.
        # The start for the fall is the 'changeframe'.
        # The end must be the end of all animation, not the given EndFrame.
        # Otherwise some parts can get caught in midair!
        # When hits ground (y=0) stop.
        for NewBubbleName in self.BubbleList.keys():
            Bubble = scene.Actor(NewBubbleName)
            ChangeFrame = self.BubbleList[NewBubbleName] # Get the point where it needs to start falling
            for F in range(ChangeFrame+1,scene.NumFrames()):
                scene.SetFrame(F-1) # Change to 1-based at last moment
                # Get existing Y tran first: if <=0, don't move at all.
                Ytran =  Bubble.ParameterByCode(poser.kParmCodeYTRAN)
                OldYVal = Ytran.Value()
                if OldYVal != DestY:
                    Xtran =  Bubble.ParameterByCode(poser.kParmCodeXTRAN)
                    Ztran =  Bubble.ParameterByCode(poser.kParmCodeZTRAN)
                    # Move all dims by Drift, and force the Y up or down by Increment.
                    if Drift > 0.0: # Avoid making keyframes in the pz3 when not needed
                        NewXVal = Xtran.Value() + ((random.random() * 2.0 - 1.0) * Drift)
                        NewZVal = Ztran.Value() + ((random.random() * 2.0 - 1.0) * Drift)
                        Xtran.SetValue(NewXVal)
                        Ztran.SetValue(NewZVal)
                    OldYVal = Ytran.Value()
                    if OldYVal < DestY: # Move upward
                        NewYVal = Ytran.Value() + Speed
                        if NewYVal > DestY: NewYVal = DestY # Force equal if beyond
                    elif OldYVal > DestY: # Move downward
                        NewYVal = Ytran.Value() - Speed
                        if NewYVal < DestY: NewYVal = DestY # Force equal if beyond
                    Ytran.SetValue(NewYVal)

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents(1)
        root.lift()
        root.after(250, self.Update)

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

root = Tk()
app = App(root, '')
app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
