#-----------------------------------------------------------------------------
# Bubbles v3
# 3/31/04
# Copyright (c) 2004 David G. Drumright (ockham)
#-----------------------------------------------------------------------------
# The initially selected object is the 'wand'.  In the script, you pick the
# bubble and set the random and movement parms.  After activation, the
# wand will leave copies of the bubble behind.  The copies will move according
# to the random choices.
#-----------------------------------------------------------------------------

from Tkinter import *
root = Tk()
import tkFileDialog
import poser
import random
import string
import pickle
import os
random.seed()
random.seed()
scene = poser.Scene()
BasePathUncut=poser.AppLocation()
BasePath=BasePathUncut[:(-len('poser.exe'))]
BubblePath=os.path.normpath(BasePath+'/Runtime/Libraries/props/*.PP2')
FPS=float(scene.FramesPerSecond())
SPF=1.0/FPS
TotalSecs=float(scene.NumFrames())/FPS

#-----------------------------------------------------------------------------
# These have to be plain globals, not self.vars, for the sake of pickle.
global app

global HoldName
global FileName
global Parented
global StartFrame
global EndFrame
global EmitSpeed
global RandEmit
global RandSize
global RandAngle
global DestY
global FallXSpeed
global FallYSpeed
global FallZSpeed
global Drift

HoldName=''
FileName=''
Parented=0
StartFrame=1
EndFrame=scene.NumFrames()
EmitSpeed=0.0
RandEmit=0.0
RandSize=0.0
RandAngle=0.0
DestY=0.0
FallXSpeed=0.0
FallYSpeed=0.0
FallZSpeed=0.0
Drift=0.0

#-----------------------------------------------------------------------------
#-----------------------------------------------------------------------------
# Next section is TK
#------------------------------------------------------------

class App:
    def __init__(self, master, textMessage):
        global HoldName
        global FileName
        global Parented
        global StartFrame
        global EndFrame
        global EmitSpeed
        global RandEmit
        global RandSize
        global RandAngle
        global DestY
        global FallXSpeed
        global FallYSpeed
        global FallZSpeed
        global Drift
        self.master = master
        master.title("Bubbles")

        self.ParentVar=IntVar()

        #------  Lay out frames

        self.TopFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.TopFrame.grid(row=0,column=0,sticky=W+E,columnspan=2)
        self.SliderFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.SliderFrame.grid(row=1,column=0,sticky=W+E,columnspan=2)
        self.ActionFrame=Frame(self.master,borderwidth=2,relief=RIDGE)
        self.ActionFrame.grid(row=2,column=0,sticky=W+E,columnspan=2)
        # Go button and delete-last, outside frames at bottom
        self.GoButton = Button(self.master, text="Apply", bg="green",command=self.HandleGo)
        self.GoButton.grid(row=3, column=0)
        self.DeleteLastButton = Button(self.master, text="Delete last set", bg="red",command=self.DeleteLast)
        self.DeleteLastButton.grid(row=3, column=1)

        #------ Items in Top Frame

        self.StatusEntry = Entry(self.TopFrame,width=40)
        self.StatusEntry.grid(row=0,column=0,columnspan=2)

        self.SelectPartButton = Button(self.TopFrame, text="Select bubble", command=self.HandleSelect)
        self.SelectPartButton.grid(row=1, column=0)

        self.ParentLabel = Label(self.TopFrame, text="Parented:")
        self.ParentLabel.grid(row=2, column=0)
        self.ParentCheck = Checkbutton(self.TopFrame,variable=self.ParentVar)
        self.ParentCheck.grid(row=2, column=1)

        self.StartLabel=Label(self.TopFrame, text="Start frame:")
        self.StartLabel.grid( row=3,column=0)
        self.StartEntry=Entry(self.TopFrame, width=6)
        self.StartEntry.grid(row=3,column=1)

        self.EndLabel=Label(self.TopFrame, text="End frame:")
        self.EndLabel.grid( row=4,column=0)
        self.EndEntry=Entry(self.TopFrame, width=6)
        self.EndEntry.grid(row=4,column=1)

        #------ Items in Slider Frame

        self.EmitSpeedLabel=Label(self.SliderFrame, text="Seconds between emissions:")
        self.EmitSpeedLabel.grid(row=0,column=0,sticky=W)
        self.EmitSpeedSlider=Scale(self.SliderFrame, from_ = SPF, to = TotalSecs,orient=HORIZONTAL,resolution=0.1)
        # Note that this starts and defaults at max possible, namely one per frame.
        self.EmitSpeedSlider.grid(row=0, column=1,columnspan=2)

        self.RandEmitLabel=Label(self.SliderFrame, text="Randomness of emission:")
        self.RandEmitLabel.grid(row=1,column=0,sticky=W)
        self.RandEmitSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        # Note that this starts and defaults at 1 per second.
        self.RandEmitSlider.grid(row=1, column=1,columnspan=2)

        self.RandSizeLabel=Label(self.SliderFrame, text="Randomness of size:")
        self.RandSizeLabel.grid( row=2,column=0,sticky=W)
        self.RandSizeSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandSizeSlider.grid(row=2,column=1,columnspan=2)

        self.RandAngleLabel=Label(self.SliderFrame, text="Randomness of angle:")
        self.RandAngleLabel.grid( row=3,column=0,sticky=W)
        self.RandAngleSlider=Scale(self.SliderFrame, from_ = 0, to = 100,orient=HORIZONTAL)
        self.RandAngleSlider.grid(row=3,column=1,columnspan=2)

        #------ Items in Action Frame

        self.DestLabel=Label(self.ActionFrame, text="Stop height:")
        self.DestLabel.grid( row=0,column=0,sticky=W)
        self.DestEntry=Entry(self.ActionFrame, width=6)
        self.DestEntry.grid(row=0,column=1)

        self.FallYLabel=Label(self.ActionFrame, text="Y speed:")
        self.FallYLabel.grid( row=1,column=0,sticky=W)
        self.FallYSlider=Scale(self.ActionFrame, from_ = -10.0, to = 10.0,orient=HORIZONTAL)
        self.FallYSlider.grid(row=1,column=1)

        self.FallXLabel=Label(self.ActionFrame, text="X speed:")
        self.FallXLabel.grid( row=2,column=0,sticky=W)
        self.FallXSlider=Scale(self.ActionFrame, from_ = -10.0, to = 10.0,orient=HORIZONTAL)
        self.FallXSlider.grid(row=2,column=1)

        self.FallZLabel=Label(self.ActionFrame, text="Z speed:")
        self.FallZLabel.grid( row=3,column=0,sticky=W)
        self.FallZSlider=Scale(self.ActionFrame, from_ =-10.0, to = 10.0,orient=HORIZONTAL)
        self.FallZSlider.grid(row=3,column=1)

        self.DriftLabel=Label(self.ActionFrame, text="Drift:")
        self.DriftLabel.grid( row=4,column=0,sticky=W)
        self.DriftSlider=Scale(self.ActionFrame, from_ = 0.0, to = 10.0,orient=HORIZONTAL)
        self.DriftSlider.grid(row=4,column=1)

        #------ Get previous settings if any, and put them in the sliders
        self.Restore() # Fill in globals

        # Set these here, not inside Restore; this setting will thus use the orig defaults
        # if Restore didn't find a dat file.
        self.ParentVar.set(Parented)
        self.StartEntry.insert(0,str(StartFrame))
        self.EndEntry.insert(0,str(EndFrame))
        self.EmitSpeedSlider.set(EmitSpeed)
        self.RandEmitSlider.set(RandEmit)
        self.RandSizeSlider.set(RandSize)
        self.RandAngleSlider.set(RandAngle)
        self.DestEntry.insert(0,str(DestY))
        self.FallXSlider.set(FallXSpeed)
        self.FallYSlider.set(FallYSpeed)
        self.FallZSlider.set(FallZSpeed)
        self.DriftSlider.set(Drift)

        #------ Set up the Wand.
        # The Wand must be the selected item.
        self.WandItem = scene.CurrentActor()

        # If it's part of a figure, get the figure as well.
        self.WandFig = self.WandItem.ItsFigure()
        if (not self.WandItem.IsBodyPart()) and (not self.WandItem.IsProp()):
            self.ShowStatus("Wand prop not a good type.  Cancel and restart.")
        else:
            self.ShowStatus("First select the bubble prop.")

        # Set up close action
        self.master.protocol('WM_DELETE_WINDOW', self.HandleCancel)

    # - - - - - - - - - - - - - - - - - -

    def ShowStatus(self,S): # Just saving repetition
        self.StatusEntry.delete(0,END)
        self.StatusEntry.insert(0,S)
        self.StatusEntry.update_idletasks()

    # - - - - - - - - - - - - - - - - - -

    def HandleSelect(self): # Get the filename for the bubble.
        # Pop up file box to get a Part.
        global FileName
        fn=tkFileDialog.askopenfilename(initialfile=BubblePath, filetypes=[("Props", "PP2")] )
        if not fn:
            self.ShowStatus("No file selected. Try again.")
            return
        FileName=os.path.normpath(fn)
        self.ShowStatus("Now set value sliders, then hit Go.")

    # - - - - - - - - - - - - - - - - - -

    def HandleGo(self):
        if not self.SetupAll():
            self.ShowStatus("No bubble selected.")
        else:
            self.Fall()
            self.Preserve()  # Hold all for next time
            self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def Preserve(self): # Hold the last set of variables to make trial and error easier.
        # Has to use global vars, not self.vars. Pickle doesn't like TK.
        global HoldName
        global FileName
        global Parented
        global StartFrame
        global EndFrame
        global EmitSpeed
        global RandEmit
        global RandSize
        global RandAngle
        global DestY
        global FallXSpeed
        global FallYSpeed
        global FallZSpeed
        global Drift

        PreserveList=[]

        PreserveList.append(HoldName)
        PreserveList.append(FileName)
        PreserveList.append(Parented)
        PreserveList.append(StartFrame)
        PreserveList.append(EndFrame)
        PreserveList.append(EmitSpeed)
        PreserveList.append(RandEmit)
        PreserveList.append(RandSize)
        PreserveList.append(RandAngle)
        PreserveList.append(DestY)
        PreserveList.append(FallXSpeed)
        PreserveList.append(FallYSpeed)
        PreserveList.append(FallZSpeed)
        PreserveList.append(Drift)

        fpDat=open("bubbles.dat","wt")
        pickle.dump(PreserveList,fpDat)
        fpDat.close()

    #-----------------------------------------------------------------------------

    def Restore(self): # Pick up the last set of variables to make trial and error easier.
        # Has to use global vars, not self.vars. Pickle doesn't like TK.
        global HoldName
        global FileName
        global Parented
        global StartFrame
        global EndFrame
        global EmitSpeed
        global RandEmit
        global RandSize
        global RandAngle
        global DestY
        global FallXSpeed
        global FallYSpeed
        global FallZSpeed
        global Drift

        fpDat=0
        try:
            fpDat=open("bubbles.dat","rt")
        except:
            return # Just do nothing

        PreserveList=pickle.load(fpDat)
        fpDat.close()

        HoldName=  PreserveList[ 0]
        FileName=  PreserveList[ 1]
        Parented=  PreserveList[ 2]
        StartFrame=PreserveList[ 3]
        EndFrame=  PreserveList[ 4]
        EmitSpeed= PreserveList[ 5]
        RandEmit=  PreserveList[ 6]
        RandSize=  PreserveList[ 7]
        RandAngle= PreserveList[ 8]
        DestY=     PreserveList[ 9]
        FallXSpeed=PreserveList[10]
        FallYSpeed=PreserveList[11]
        FallZSpeed=PreserveList[12]
        Drift=     PreserveList[13]


    # - - - - - - - - - - - - - - - - - -

    def SetupAll(self):
        # Returns 1 if good, 0 if no bubble prop.
        global EmitSpeed
        global RandEmit
        global RandSize
        global RandAngle
        global StartFrame
        global EndFrame
        global HoldName
        global Parented

        if (not FileName) or (FileName==''):
            self.ShowStatus("You need to select a bubble prop first.")
            return 0
        self.ShowStatus("Working....")
        # Get the sliders we need here.
        EmitSpeed = self.EmitSpeedSlider.get()
        if EmitSpeed < 0.0001: return
        RandEmit= self.RandEmitSlider.get()

        Parented=self.ParentVar.get()

        # Figure the interval from the speed.  Speed is seconds per bubble;
        # we need frames per bubble.   (F/S) * (S/B) = F/B.
        EmitInterval = int(float(scene.FramesPerSecond()) * EmitSpeed)
        if EmitInterval < 1: EmitInterval = 1

        RandSize = float(self.RandSizeSlider.get()) / 100.0
        # For the angle, max move should be 180 degrees.
        RandAngle = float(self.RandAngleSlider.get()) * 180.0 / 100.0

        s = self.StartEntry.get()
        if s: StartFrame = int(s)
        s = self.EndEntry.get()
        if s: EndFrame = int(s)

        # First loop.
        # Go through the frames, watching the Endpoint of the 'wand'.
        # At intervals given by the EmitSpeed, produce a copy of the 'bubble',
        # at this location, and hold the frame so we can use the second
        # loop to make it appear (scale=1) at the appropriate frame.

        self.BubbleList = {} # Dict: key is name, value is frame.

        # Note that Start and End frame affect the range of creation as well as the
        # range of fall.

        # Always create one copy at StartFrame. After that, we may create several
        # more depending on how the Interval fits.
        F = StartFrame
        while F < EndFrame:  # Loop this way so we can randomize the interval more easily

            scene.SetFrame(F-1)

            scene.DrawAll()

            # Treat the Origin of the WandItem, converted to world coords,
            # as the point of emission.  This lets you pick the exact point
            # conveniently by moving the Origin.
            (x,y,z) = self.WandItem.Origin()
            ( (a,b,c,d) , (e,f,g,h), (i,j,k,l), (m,n,p,q) ) = self.WandItem.WorldMatrix()
            w=1.0
            x1 = x*a + y*e + z*i + w*m
            y1 = x*b + y*f + z*j + w*n
            z1 = x*c + y*g + z*k + w*p
            w1 = x*d + y*h + z*l + w*q
            EmitX = x1/w1
            EmitY = y1/w1
            EmitZ = z1/w1

            # Go back to 0 and make a copy of the Bubble.
            # Reminder: We do a frame-switch here
            # because the dot needs to be at this location from the start, even though
            # we are acting at frame F.
            scene.SetFrame(0)
            scene.LoadLibraryProp(FileName)
            # The new prop will be currentactor at this moment.
            NewBubble = scene.CurrentActor()

            # For the preserve and undo, hold the base name of the bubbles.
            FullName=NewBubble.Name()
            Pos=string.find(FullName,'_')
            if Pos >=0: HoldName=FullName[:Pos]
            else:       HoldName=FullName

            # Control its scale randomly. Reminder: the original will be the same
            # every time, but this is easier than trying to pick up a "sample"
            # before the rest.

            YrotParm=NewBubble.ParameterByCode(poser.kParmCodeYROT)
            XrotParm=NewBubble.ParameterByCode(poser.kParmCodeXROT)
            XtranParm=NewBubble.ParameterByCode(poser.kParmCodeXTRAN)
            YtranParm=NewBubble.ParameterByCode(poser.kParmCodeYTRAN)
            ZtranParm=NewBubble.ParameterByCode(poser.kParmCodeZTRAN)

            Yrot=YrotParm.Value() + (random.random()*2.0-1.0)*RandAngle
            Xrot=XrotParm.Value() + (random.random()*2.0-1.0)*RandAngle

            XtranParm.SetValue(EmitX)
            YtranParm.SetValue(EmitY)
            ZtranParm.SetValue(EmitZ)
            YrotParm.SetValue(Yrot)
            XrotParm.SetValue(Xrot)

            # Record the name and the frame when we made it
            self.BubbleList[FullName]=F
            Interval = EmitInterval + (random.random()*2.0-1.0)*RandEmit
            if Interval<1:Interval=1
            F = F + Interval # Reminder: this is a While; will quit when F > EndFrame.
            # end While

        scene.DrawAll()

        # Second loop: go through the created Bubbles. For each,
        # scale to zero from start, and jump to full scale from the frame where created.
        self.ShowStatus("Yes, I'm still working... Just wait till I say Done.")
        for NewBubbleName in self.BubbleList.keys():
            ChangeFrame = self.BubbleList[NewBubbleName] # Get the point where it needs to jump (appear)

            if ChangeFrame < StartFrame: continue # Skip bad ones
            elif ChangeFrame > EndFrame: continue # Skip bad ones

            NewBubble = scene.Actor(NewBubbleName)
            XscaleParm=NewBubble.ParameterByCode(poser.kParmCodeXSCALE)
            YscaleParm=NewBubble.ParameterByCode(poser.kParmCodeYSCALE)
            ZscaleParm=NewBubble.ParameterByCode(poser.kParmCodeZSCALE)
            AscaleParm=NewBubble.ParameterByCode(poser.kParmCodeASCALE)
            OrigXscale=XscaleParm.Value()
            OrigYscale=YscaleParm.Value()
            OrigZscale=ZscaleParm.Value()
            OrigAscale=AscaleParm.Value()

            # If the Rand Size slider is left at zero, these vals are just same as Orig.
            RandXscale=OrigXscale + (random.random()*2.0-1.0) * RandSize
            RandYscale=OrigYscale + (random.random()*2.0-1.0) * RandSize
            RandZscale=OrigZscale + (random.random()*2.0-1.0) * RandSize

            # Up to ChangeFrame, set scale 0.
            for F in range(1,ChangeFrame): # From 0, not from StartFrame.
                scene.SetFrame(F-1)
                XscaleParm.SetValue(OrigXscale)
                YscaleParm.SetValue(OrigYscale)
                ZscaleParm.SetValue(OrigZscale)
                AscaleParm.SetValue(0.0) # Here's what makes the object 'invisible' before emission
            # From ChangeFrame to end, set main scale to orig.
            for F in range(ChangeFrame,EndFrame):
                scene.SetFrame(F-1)
                XscaleParm.SetValue(RandXscale)
                YscaleParm.SetValue(RandYscale)
                ZscaleParm.SetValue(RandZscale)
                AscaleParm.SetValue(OrigAscale)
            if Parented:
                NewBubble.SetParent(self.WandItem,1,0)

        scene.DrawAll()
        return 1

    # - - - - - - - - - - - - - - - - - -

    def Fall(self): # Acts only if speed has been set
        global FallXSpeed
        global FallYSpeed
        global FallZSpeed
        global DestY
        global Drift
        global StartFrame
        global EndFrame

        FallXSpeed = self.FallXSlider.get() / 1000.0
        FallYSpeed = self.FallYSlider.get() / 1000.0
        FallZSpeed = self.FallZSlider.get() / 1000.0

        s=self.DestEntry.get()
        if s: DestY = float(s)
        else: DestY = 0.0
        # DestY is used only when not parented: too confusing otherwise.
        # This may be above or below the original position.
        # If above, we increment Y upward and stop when Y >= DestY.
        # If below, we increment Y dowward and stop when Y <= DestY.
        # To avoid slippery comparisons, we force Y = DestY at the stop point.

        # Fetch drift percent:
        Drift = self.DriftSlider.get() / 500.0
        # No guarding needed here; if 0, it will just naturally
        # leave no random movement.

        StartFrame=1
        EndFrame=scene.NumFrames()
        s = self.StartEntry.get()
        if s: StartFrame = int(s)
        s = self.EndEntry.get()
        if s: EndFrame = int(s)
        if (EndFrame - StartFrame) < 2:
            self.ShowStatus("StartFrame and EndFrame are improper.")
            return

        # From startframe to endframe, move each particle down by Increment, and
        # sideways by Drift.
        # The start for the fall is the 'changeframe'.
        # The end must be the end of all animation, not the given EndFrame.
        # Otherwise some parts can get caught in midair!
        # When hits ground (y=0) stop.
        for NewBubbleName in self.BubbleList.keys():
            Bubble = scene.Actor(NewBubbleName)
            ChangeFrame = self.BubbleList[NewBubbleName] # Get the point where it needs to start falling

            if ChangeFrame < StartFrame: continue # Skip those that won't fall anyway
            elif ChangeFrame >= EndFrame: continue
            Xtran = Bubble.ParameterByCode(poser.kParmCodeXTRAN)
            Ytran = Bubble.ParameterByCode(poser.kParmCodeYTRAN)
            Ztran = Bubble.ParameterByCode(poser.kParmCodeZTRAN)

            # For each bubble, run from ChangeFrame to total end.
            for F in range(ChangeFrame+1,scene.NumFrames()): # Run to absolute end, not EndFrame
                scene.SetFrame(F-1) # Change to 1-based at last moment
                # Get existing Y tran first: if <=0, don't move at all.

                if (not Parented) and (Ytran.Value() == DestY): break # When not parented, stop at Dest.

                # Move all dims only when the sliders are nonzero, to avoid making
                # unneeded 'k' groups in the PZ3.

                if FallXSpeed != 0.0:
                    NewXVal = Xtran.Value() + FallXSpeed
                    Xtran.SetValue(NewXVal)

                if FallZSpeed != 0.0:
                    NewZVal = Ztran.Value() + FallZSpeed
                    Ztran.SetValue(NewZVal)

                if FallYSpeed != 0.0:
                    NewYVal = Ytran.Value() + FallYSpeed
                    # When parented we skip the dest stuff.
                    if not Parented:
                        if   (FallYSpeed > 0.0) and (NewYVal > DestY) : NewYVal = DestY # Force equal if beyond
                        elif (FallYSpeed < 0.0) and (NewYVal < DestY) : NewYVal = DestY # Force equal if beyond
                    Ytran.SetValue(NewYVal)

                if Drift > 0.0:
                    # Note that we use the actual Value() here, so that we add onto the above changes
                    NewXVal = Xtran.Value() + ((random.random() * 2.0 - 1.0) * Drift)
                    NewZVal = Ztran.Value() + ((random.random() * 2.0 - 1.0) * Drift)
                    Xtran.SetValue(NewXVal)
                    Ztran.SetValue(NewZVal)

        self.ShowStatus("Done.")

    # - - - - - - - - - - - - - - - - - -

    def DeleteLast(self):
        # Take the HoldName from last Restore. If anything in the scene
        # fits that name, delete it.
        for OneActor in scene.Actors():
            Name=OneActor.Name()
            if string.count(Name,HoldName):
                scene.SelectActor(OneActor)
                scene.DeleteCurrentProp()

    # - - - - - - - - - - - - - - - - - -

    def HandleCancel(self):
        self.master.destroy()

    # - - - - - - - - - - - - - - - - - -

    def Update(self):
        if scene: scene.ProcessSomeEvents(1)
        root.lift()
        root.after(250, self.Update)

#------------------------------------------------------------
# Activate the loop
#------------------------------------------------------------

app = App(root, '')
#app.Update()
root.mainloop()

#------------------------------------------------------------
# End TK loop.
#-----------------------------------------------------------------------------
